use super::{
    parse::{EntityKey, Verb},
    Autogenerate, CLIError, EntityInterface,
};
use crate::{
    query::{Insertable, Queryable, RelationInterface},
    schema::{
        datum::{ConcreteDatumList, Datum, DatumDiscriminatorRef},
        entity::{Entity, EntityID, EntityPart, EntityPartList, EntityPartVisitor},
        relation::{Relation, RelationDomain, RelationMap, RelationRange},
    },
    ConnectionLease,
};

// helper alias for later
type UniqueList<E> = <<E as Entity>::Keys as EntityPartList>::DatumList;

impl<EI: EntityInterface> Autogenerate<EI> {
    /// Execute the parsed command.
    pub fn perform(
        self,
        ctx: &EI::Context,
        lease: &mut ConnectionLease,
        query_ctx: impl Queryable<EntityOutput = EI::Entity> + Insertable<EI::Entity>,
    ) -> Result<(), EI::Error> {
        match self.verb {
            Verb::Attach {
                local_keys,
                relation,
                remote_keys,
            } => {
                let local_keys = EntityKey::to_string_vec::<EI>(&local_keys, ctx);
                let remote_keys = EntityKey::to_string_vec::<EI>(&remote_keys, ctx);
                let outer_obj = query_ctx
                    .keyed(
                        UniqueList::<EI::Entity>::build_equivalent(
                            local_keys.iter().map(String::as_str),
                        )
                        .unwrap(),
                    )
                    .get(lease)?
                    .ok_or(<EI::Error>::no_such_entity(
                        EI::Entity::entity_name(),
                        local_keys
                            .iter()
                            .cloned()
                            .reduce(|a, b| format!("{},{}", a, b))
                            .unwrap()
                            .to_string(),
                    ))?;

                let mut attacher = Attacher {
                    do_attach: true,
                    relation: relation.as_str(),
                    lease,
                    remote_keys,
                    err: None,
                    _ghost: Default::default(),
                };
                outer_obj.accept_part_visitor_ref(&mut attacher);

                if let Some(err) = attacher.err {
                    return Err(err);
                }
            },
            Verb::Delete(keys) => {
                let keys = EntityKey::to_string_vec::<EI>(&keys, ctx);
                query_ctx
                    .keyed(
                        UniqueList::<EI::Entity>::build_equivalent(keys.iter().map(String::as_str))
                            .unwrap(),
                    )
                    .delete(lease)?;
            },
            Verb::Detach {
                local_keys,
                relation,
                remote_keys,
            } => {
                let local_keys = EntityKey::to_string_vec::<EI>(&local_keys, ctx);
                let remote_keys = EntityKey::to_string_vec::<EI>(&remote_keys, ctx);
                let outer_obj = query_ctx
                    .keyed(
                        UniqueList::<EI::Entity>::build_equivalent(
                            local_keys.iter().map(String::as_str),
                        )
                        .unwrap(),
                    )
                    .get(lease)?
                    .ok_or(<EI::Error>::no_such_entity(
                        EI::Entity::entity_name(),
                        local_keys
                            .iter()
                            .cloned()
                            .reduce(|a, b| format!("{},{}", a, b))
                            .unwrap()
                            .to_string(),
                    ))?;

                let mut attacher = Attacher {
                    do_attach: false,
                    relation: relation.as_str(),
                    lease,
                    remote_keys,
                    err: None,
                    _ghost: Default::default(),
                };
                outer_obj.accept_part_visitor_ref(&mut attacher);

                if let Some(err) = attacher.err {
                    return Err(err);
                }
            },
            Verb::ListAll => {
                println!(
                    "Listing all {}(s): ({})",
                    EI::Entity::entity_name(),
                    query_ctx.clone().count(lease)?
                );
                for obj in query_ctx.get(lease)?.into_iter() {
                    println!(
                        " - {}",
                        EI::summarize(&obj).unwrap_or_else(|| format!("{:?}", obj))
                    );
                }
            },
            Verb::Inspect(keys) => {
                let keys = EntityKey::to_string_vec::<EI>(&keys, ctx);
                let obj = query_ctx
                    .keyed(
                        UniqueList::<EI::Entity>::build_equivalent(keys.iter().map(String::as_str))
                            .unwrap(),
                    )
                    .get(lease)?
                    .ok_or(<EI::Error>::no_such_entity(
                        EI::Entity::entity_name(),
                        keys.iter()
                            .cloned()
                            .reduce(|a, b| format!("{},{}", a, b))
                            .unwrap()
                            .to_string(),
                    ))?;
                println!("{:#?}", obj.as_ref());

                fn inspect_ai<AI: RelationInterface>(
                    name: &'static str,
                    lease: &mut ConnectionLease<'_>,
                    ai: &AI,
                ) {
                    println!("{}: ({})", name, ai.count(lease).unwrap());
                    for a in ai.get(lease).expect("couldn't get object relations") {
                        println!("[#{:3}]: {:?}", a.id().into_raw(), a.wrapped());
                    }
                }

                struct RelationFieldWalker<'r, 'l: 'r, E: Entity>(
                    &'r mut ConnectionLease<'l>,
                    std::marker::PhantomData<E>,
                );
                impl<'r, 'l: 'r, E: Entity> EntityPartVisitor for RelationFieldWalker<'r, 'l, E> {
                    type Entity = E;
                    fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                        struct Discriminator<'r, 'l: 'r>(&'r mut ConnectionLease<'l>, &'static str);

                        impl<'r, 'l> DatumDiscriminatorRef for Discriminator<'r, 'l> {
                            fn visit_serialized<
                                T: serde::Serialize + serde::de::DeserializeOwned,
                            >(
                                &mut self,
                                _: &T,
                            ) {
                            }
                            fn visit_bare_field<T: Datum>(&mut self, _: &T) {}
                            fn visit_entity_id<E: Entity>(&mut self, _: &E::ID) {}
                            fn visit_relation_map<E: Entity>(&mut self, amap: &RelationMap<E>) {
                                inspect_ai(self.1, self.0, amap);
                            }
                            fn visit_relation_domain<R: Relation>(
                                &mut self,
                                adomain: &RelationDomain<R>,
                            ) {
                                inspect_ai(self.1, self.0, adomain);
                            }
                            fn visit_relation_range<R: Relation>(
                                &mut self,
                                arange: &RelationRange<R>,
                            ) {
                                inspect_ai(self.1, self.0, arange);
                            }
                        }

                        datum.accept_discriminator_ref(&mut Discriminator(self.0, EP::part_name()));
                    }
                }

                obj.accept_part_visitor_ref(&mut RelationFieldWalker(lease, Default::default()));
            },
            Verb::Custom(custom) => {
                EI::run_custom(ctx, custom, lease, query_ctx)?;
            },
        }

        Ok(())
    }
}

/// helper type for attach and detach verbs
struct Attacher<'r, 'l: 'r, Error: CLIError, E: Entity> {
    do_attach: bool,
    relation: &'r str,
    lease: &'r mut ConnectionLease<'l>,
    remote_keys: Vec<String>,
    err: Option<Error>,
    _ghost: std::marker::PhantomData<E>,
}

impl<'r, 'l: 'r, Error: CLIError, OE: Entity> Attacher<'r, 'l, Error, OE> {
    fn do_operation<E: Entity>(&mut self, map: &impl RelationInterface<RemoteEntity = E>) {
        match map
            .query_all()
            .keyed(
                UniqueList::<E>::build_equivalent(self.remote_keys.iter().map(String::as_str))
                    .unwrap(),
            )
            .get(self.lease)
        {
            Ok(Some(obj)) => {
                if self.do_attach {
                    self.err = map.connect_to(self.lease, obj.id()).err().map(Into::into);
                } else {
                    self.err = map
                        .disconnect_from(self.lease, obj.id())
                        .err()
                        .map(Into::into);
                }
            },
            Ok(None) => {
                self.err = Some(Error::no_such_entity(
                    E::entity_name(),
                    self.remote_keys
                        .iter()
                        .cloned()
                        .reduce(|a, b| format!("{},{}", a, b))
                        .unwrap_or_default()
                        .to_string(),
                ));
            },
            Err(e) => {
                self.err = Some(e.into());
            },
        }
    }
}

impl<'r, 'l: 'r, Error: CLIError, E: Entity> EntityPartVisitor for Attacher<'r, 'l, Error, E> {
    type Entity = E;
    fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
        if EP::part_name() != self.relation {
            return;
        }

        datum.accept_discriminator_ref(self);
    }
}

impl<'r, 'l: 'r, Error: CLIError, OE: Entity> DatumDiscriminatorRef
    for Attacher<'r, 'l, Error, OE>
{
    fn visit_entity_id<E: Entity>(&mut self, _: &E::ID) {
        unreachable!()
    }
    fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self, _: &T) {
        unreachable!()
    }
    fn visit_bare_field<T: Datum>(&mut self, _: &T) {
        unreachable!()
    }

    fn visit_relation_map<E: Entity>(&mut self, map: &RelationMap<E>) {
        self.do_operation(map);
    }
    fn visit_relation_range<R: Relation>(&mut self, map: &RelationRange<R>) {
        self.do_operation(map);
    }
    fn visit_relation_domain<R: Relation>(&mut self, map: &RelationDomain<R>) {
        self.do_operation(map);
    }
}
