//! Database query interface.
//!
//!

use itertools::Itertools;

use crate::{
    db::{ConnectionLease, StatementContext, Transaction},
    schema::{
        datum::{Datum, QueryEquivalent, QueryEquivalentList},
        entity::{Entity, EntityID, EntityPart, EntityPartList},
        index::Index,
        relation::{LocalSide, RelationData},
        Stored,
    },
    DBResult, Error,
};

use std::collections::HashMap;
use std::hash::{Hash, Hasher};

pub(crate) mod base_queries;
pub(crate) mod components;
pub(crate) mod containers;
use containers::*;

#[derive(Debug, Clone)]
pub(crate) enum QueryPartData<'l> {
    Owned(String),
    Borrowed(&'l str),
}

impl<'l> std::fmt::Display for QueryPartData<'l> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Owned(s) => f.write_str(s),
            Self::Borrowed(s) => f.write_str(s),
        }
    }
}

impl<'l> From<String> for QueryPartData<'l> {
    fn from(value: String) -> Self {
        Self::Owned(value)
    }
}

impl<'l> From<&'l str> for QueryPartData<'l> {
    fn from(value: &'l str) -> Self {
        Self::Borrowed(value)
    }
}

#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq)]
pub(crate) enum QueryPart {
    Root,
    Columns,
    From,
    Set,
    Join,
    Where,
    Trailing,
}

/// Represents a SQL query built during runtime.
#[derive(Debug)]
pub struct Query<'l> {
    parts: HashMap<QueryPart, Vec<QueryPartData<'l>>>,
}

impl<'l> Query<'l> {
    pub(crate) fn new() -> Self {
        Self {
            parts: Default::default(),
        }
    }

    pub(crate) fn attach<T: Into<QueryPartData<'l>>>(mut self, qp: QueryPart, val: T) -> Self {
        self.attach_mut(qp, val.into());
        self
    }

    pub(crate) fn replace<T: Into<QueryPartData<'l>>>(mut self, qp: QueryPart, val: T) -> Self {
        self.parts.remove(&qp);
        self.attach(qp, val.into())
    }

    pub(crate) fn attach_mut<T: Into<QueryPartData<'l>>>(&mut self, qp: QueryPart, val: T) {
        self.parts.entry(qp).or_default().push(val.into());
    }

    pub(crate) fn assemble(mut self) -> String {
        let root = self.parts.remove(&QueryPart::Root).unwrap().remove(0);

        let columns_ = match self.parts.remove(&QueryPart::Columns) {
            None => String::new(),
            Some(v) => v.into_iter().join(","),
        };

        let from_ = match self.parts.remove(&QueryPart::From) {
            None => String::new(),
            Some(v) => {
                format!("FROM {}", v.into_iter().join(","))
            },
        };

        let set_ = match self.parts.remove(&QueryPart::Set) {
            None => String::new(),
            Some(v) => {
                format!("SET {}", v.into_iter().join(","))
            },
        };

        let join_ = match self.parts.remove(&QueryPart::Join) {
            None => String::new(),
            Some(v) => v
                .into_iter()
                .map(|j| format!("INNER JOIN {}", j))
                .reduce(|a, b| format!("{} {}", a, b))
                .unwrap(),
        };

        let where_ = match self.parts.remove(&QueryPart::Where) {
            None => String::new(),
            Some(v) => {
                format!("WHERE {}", v.into_iter().join(" AND "))
            },
        };

        let trailing_ = match self.parts.remove(&QueryPart::Trailing) {
            None => String::new(),
            Some(v) => v.into_iter().join(" "),
        };

        format!(
            "{} {} {} {} {} {} {}",
            root, columns_, from_, set_, join_, where_, trailing_
        )
    }
}

pub(crate) struct RelationNames {
    local_name: &'static str,
    remote_name: &'static str,
    part_name: &'static str,
    dist_name: &'static str,
    domain_name: &'static str,
    range_name: &'static str,
    local_field: &'static str,
    remote_field: &'static str,
}

impl RelationNames {
    fn collect<AI: RelationInterface>(iface: &AI) -> DBResult<RelationNames> {
        let rdata = iface.get_data()?;
        let local_name = rdata.local_name;
        let remote_name = <AI::RemoteEntity>::entity_name();
        let part_name = rdata.part_name;
        let dist_name = iface.get_distinguishing_name()?;

        let (domain_name, range_name) = match AI::SIDE {
            LocalSide::Domain => (local_name, remote_name),
            LocalSide::Range => (remote_name, local_name),
        };
        let (local_field, remote_field) = match AI::SIDE {
            LocalSide::Domain => ("domain", "range"),
            LocalSide::Range => ("range", "domain"),
        };
        Ok(Self {
            local_name,
            remote_name,
            part_name,
            dist_name,
            domain_name,
            range_name,
            local_field,
            remote_field,
        })
    }

    fn relation_name(&self) -> String {
        format!(
            "{domain_name}_{range_name}_relation_{dist_name}",
            domain_name = self.domain_name,
            range_name = self.range_name,
            dist_name = self.dist_name
        )
    }
}

fn hash_of<T: Hash>(val: T) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();
    val.hash(&mut hasher);
    hasher.finish()
}

/// Relation map generic interface trait.
pub trait RelationInterface {
    /// The type of the entity on the non-local end of the relation.
    type RemoteEntity: Entity;
    /// A static version of the current interface, used for type-based query caching.
    type StaticVersion: RelationInterface + 'static;

    #[doc(hidden)]
    fn get_data(&self) -> DBResult<&RelationData>;
    #[doc(hidden)]
    fn get_distinguishing_name(&self) -> DBResult<&'static str>;

    /// Which side is the "local" side of the relation.
    const SIDE: LocalSide;

    /// Query this entity type without the relation filter.
    fn query_all(&self) -> impl Queryable<EntityOutput = Self::RemoteEntity> {
        components::TableComponent::<Self::RemoteEntity>::new()
    }

    /// Attempt to connect the contextual instance to a remote instance.
    fn connect_to(
        &self,
        lease: &mut ConnectionLease,
        remote_id: <Self::RemoteEntity as Entity>::ID,
    ) -> DBResult<()>
    where
        Self: Sized,
    {
        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let mut txn = Transaction::new(lease, "_microrm_connect_to")?;

        base_queries::do_connect::<Self::RemoteEntity>(txn.lease(), rdata, an, remote_id)?;

        txn.commit()
    }

    /// Attempt to disconnect the contextual instance from a remote instance.
    fn disconnect_from(
        &self,
        lease: &mut ConnectionLease,
        remote_id: <Self::RemoteEntity as Entity>::ID,
    ) -> DBResult<()>
    where
        Self: Sized,
    {
        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let mut txn = Transaction::new(lease, "_microrm_disconnect_from")?;

        // second, add to the relation table
        txn.lease().with_prepared(
            hash_of(("disconnect", an.local_name, an.remote_name, an.part_name)),
            || {
                format!(
                    "delete from `{relation_name}` where `{local_field}` = ? and `{remote_field}` = ?",
                    relation_name = an.relation_name(),
                    local_field = an.local_field,
                    remote_field = an.remote_field
                )
            },
            |ctx| {
                ctx.bind(1, rdata.local_id)?;
                ctx.bind(2, remote_id.into_raw())?;

                ctx.run().map(|_| ())
            },
        )?;

        txn.commit()
    }
}

// ----------------------------------------------------------------------
// New query interface
// ----------------------------------------------------------------------

/// Represents a context in which we can insert an entity type `E`.
pub trait Insertable<E: Entity> {
    /// Insert an entity instance and return its new ID.
    fn insert(&self, lease: &mut ConnectionLease, value: E) -> DBResult<E::ID>;
    /// Insert an entity instance and return a [`Stored`] instance that can be used to synchronize
    /// its values back into the database later.
    fn insert_and_return(&self, lease: &mut ConnectionLease, value: E) -> DBResult<Stored<E>>;
}

impl<AI: RelationInterface> Insertable<AI::RemoteEntity> for AI {
    fn insert(
        &self,
        lease: &mut ConnectionLease,
        value: AI::RemoteEntity,
    ) -> DBResult<<AI::RemoteEntity as Entity>::ID>
    where
        Self: Sized,
    {
        // we're doing two things:
        // - inserting the entity into the target table
        // - adding the relation row into the relation table

        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let mut txn = Transaction::new(lease, "_microrm_insert")?;

        // so first, into the remote table
        let remote_id = base_queries::insert(txn.lease(), &value)?;
        // then the relation
        base_queries::do_connect::<AI::RemoteEntity>(txn.lease(), rdata, an, remote_id)?;

        txn.commit()?;

        Ok(remote_id)
    }

    fn insert_and_return(
        &self,
        lease: &mut ConnectionLease,
        value: AI::RemoteEntity,
    ) -> DBResult<Stored<AI::RemoteEntity>>
    where
        Self: Sized,
    {
        // we're doing two things:
        // - inserting the entity into the target table
        // - adding the relation row into the relation table

        let rdata = self.get_data()?;
        let an = RelationNames::collect::<Self>(self)?;

        let mut txn = Transaction::new(lease, "_microrm_insert_and_return")?;

        // so first, into the remote table
        let remote = base_queries::insert_and_return(txn.lease(), value)?;
        // then the relation
        base_queries::do_connect::<AI::RemoteEntity>(txn.lease(), rdata, an, remote.id())?;

        txn.commit()?;

        Ok(remote)
    }
}

/// Represents a searchable context of a given entity.
pub trait Queryable: Clone {
    /// The entity that results from a search in this context.
    type EntityOutput: Entity;
    /// How results will be provided. This is either a `Vec` or an `Option`.
    type OutputContainer: OutputContainer<Self::EntityOutput>;
    /// A `'static`-version of `Self`, used for `TypeId`-based caching.
    type StaticVersion: Queryable + 'static;

    /// True if the query result is guaranteed to be either unique or absent.
    const IS_UNIQUE: bool;

    /// Construct a concrete SQL query from this abstract type representation.
    #[doc(hidden)]
    fn build(&self) -> Query;
    /// Bind into any required placeholders to 'fill' an instance created by [`build`].
    #[doc(hidden)]
    fn bind(&self, stmt: &mut StatementContext, index: &mut i32);

    // ----------------------------------------------------------------------
    // Verbs
    // ----------------------------------------------------------------------
    /// Count all entities in the current context.
    ///
    /// Returns the number of entities.
    fn count(self, lease: &mut ConnectionLease) -> DBResult<usize>
    where
        Self: Sized,
    {
        let mut txn = Transaction::new(lease, "_microrm_count")?;
        struct CountTag;
        let out = txn.lease().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, CountTag)>(),
            || {
                self.build()
                    .replace(
                        QueryPart::Columns,
                        format!(
                            "COUNT(DISTINCT `{}`.`id`)",
                            Self::EntityOutput::entity_name()
                        ),
                    )
                    .assemble()
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                Ok(ctx
                    .run()?
                    .ok_or(Error::InternalError("no resulting rows from COUNT query"))?
                    .read::<i64>(0)? as usize)
            },
        )?;
        txn.commit()?;
        Ok(out)
    }
    /// Get all entities in the current context.
    fn get(self, lease: &mut ConnectionLease) -> DBResult<Self::OutputContainer>
    where
        Self: Sized,
    {
        let mut txn = Transaction::new(lease, "_microrm_get")?;
        struct GetTag;
        let out = txn.lease().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, GetTag)>(),
            || self.build().assemble(),
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                <Self::OutputContainer>::assemble_from(ctx)
            },
        )?;
        txn.commit()?;
        Ok(out)
    }
    /// Get IDs of all entities in the current context.
    fn get_ids(
        self,
        lease: &mut ConnectionLease,
    ) -> DBResult<<Self::OutputContainer as OutputContainer<Self::EntityOutput>>::IDContainer>
    where
        Self: Sized,
    {
        let mut txn = Transaction::new(lease, "_microrm_get_ids")?;
        struct GetIDTag;
        let out = txn.lease().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, GetIDTag)>(),
            || {
                self.build()
                    .replace(
                        QueryPart::Columns,
                        format!("`{}`.`id`", Self::EntityOutput::entity_name()),
                    )
                    .assemble()
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                <<Self::OutputContainer as OutputContainer<
                        Self::EntityOutput,
                    >>::IDContainer>::assemble_from(ctx)
            },
        )?;
        txn.commit()?;
        Ok(out)
    }
    /// Delete all entities in the current context.
    fn delete(self, lease: &mut ConnectionLease) -> DBResult<()>
    where
        Self: Sized,
    {
        let mut txn = Transaction::new(lease, "_microrm_delete")?;
        struct DeleteTag;
        txn.lease().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, DeleteTag)>(),
            || {
                format!(
                    "DELETE FROM `{}` WHERE `id` = ({})",
                    Self::EntityOutput::entity_name(),
                    self.build()
                        .replace(
                            QueryPart::Columns,
                            format!("`{}`.`id`", Self::EntityOutput::entity_name())
                        )
                        .assemble()
                )
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                ctx.run()?;
                Ok(())
            },
        )?;
        txn.commit()
    }

    /// Delete all entities in the current context and return them
    fn remove(self, lease: &mut ConnectionLease) -> DBResult<Self::OutputContainer>
    where
        Self: Sized,
    {
        let mut txn = Transaction::new(lease, "_microrm_remove")?;
        struct DeleteTag;
        let out = txn.lease().with_prepared(
            std::any::TypeId::of::<(Self::StaticVersion, DeleteTag)>(),
            || {
                format!(
                    "DELETE FROM `{entity}` WHERE `id` = ({subquery}) RETURNING *",
                    entity = Self::EntityOutput::entity_name(),
                    subquery = self
                        .build()
                        .replace(
                            QueryPart::Columns,
                            format!("`{}`.`id`", Self::EntityOutput::entity_name())
                        )
                        .assemble()
                )
            },
            |mut ctx| {
                // starting index is 1
                let mut index = 1;
                self.bind(&mut ctx, &mut index);

                <Self::OutputContainer>::assemble_from(ctx)
            },
        )?;
        txn.commit()?;
        Ok(out)
    }

    // ----------------------------------------------------------------------
    // Filtering methods
    // ----------------------------------------------------------------------
    /// Filter using the keying index on the entity.
    fn keyed(
        self,
        values: impl QueryEquivalentList<
            <<Self::EntityOutput as Entity>::Keys as EntityPartList>::DatumList,
        >,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<Stored<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::IndexComponent::<_, _, <Self::EntityOutput as Entity>::Keys, _>::new(
            self, values,
        )
    }

    /// Filter using an arbitrary index on the entity.
    fn indexed<EPL: EntityPartList<Entity = Self::EntityOutput>>(
        self,
        _index: &Index<Self::EntityOutput, EPL>,
        values: impl QueryEquivalentList<EPL::DatumList>,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<Stored<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::IndexComponent::<_, _, EPL, _>::new(self, values)
    }

    /// Filter using an arbitrary column on the entity.
    fn with<EP: EntityPart<Entity = Self::EntityOutput>>(
        self,
        part: EP,
        value: impl QueryEquivalent<EP::Datum>,
    ) -> impl Queryable<EntityOutput = Self::EntityOutput, OutputContainer = Self::OutputContainer>
    where
        Self: Sized,
    {
        components::WithComponent::new(self, part, value)
    }

    /// Filter exactly on an entity ID.
    fn with_id(
        self,
        id: <Self::EntityOutput as Entity>::ID,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<Stored<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        self.with(<Self::EntityOutput as Entity>::IDPart::default(), id)
            .first()
    }

    /// Ask to return at most a single result.
    fn first(
        self,
    ) -> impl Queryable<
        EntityOutput = Self::EntityOutput,
        OutputContainer = Option<Stored<Self::EntityOutput>>,
    >
    where
        Self: Sized,
    {
        components::SingleComponent::new(self)
    }

    // ----------------------------------------------------------------------
    // Relation-following and joining methods
    // ----------------------------------------------------------------------
    /// Join based on an existing relation.
    fn join<
        AD: RelationInterface + Datum,
        EP: EntityPart<Entity = Self::EntityOutput, Datum = AD>,
    >(
        self,
        part: EP,
    ) -> impl Queryable<EntityOutput = AD::RemoteEntity, OutputContainer = Vec<Stored<AD::RemoteEntity>>>
    where
        Self: Sized,
    {
        components::JoinComponent::<AD::RemoteEntity, Self::EntityOutput, _, Self>::new(self, part)
    }

    /// Follow a foreign key.
    fn foreign<EP: EntityPart<Entity = Self::EntityOutput>>(
        self,
        part: EP,
    ) -> impl Queryable<
        EntityOutput = <EP::Datum as EntityID>::Entity,
        OutputContainer = <Self::OutputContainer as OutputContainer<
            Self::EntityOutput,
        >>::ReplacedEntity<<EP::Datum as EntityID>::Entity>,
    >
    where
        Self: Sized,
        EP::Datum: EntityID,
    {
        components::ForeignComponent::<_, EP, Self>::new(self, part)
    }
}

// Generic implementation for all relation specification types
impl<'a, AI: RelationInterface> Queryable for &'a AI {
    type EntityOutput = AI::RemoteEntity;
    type OutputContainer = Vec<Stored<AI::RemoteEntity>>;
    type StaticVersion = &'static AI::StaticVersion;

    const IS_UNIQUE: bool = false;

    fn build(&self) -> Query {
        let anames = RelationNames::collect(*self).unwrap();
        let relation_name = anames.relation_name();
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT")
            .attach(QueryPart::Columns, format!("`{}`.*", anames.remote_name))
            .attach(QueryPart::From, format!("`{}`", relation_name))
            .attach(
                QueryPart::Join,
                format!(
                    "`{}` ON `{}`.`id` = `{}`.`{}`",
                    anames.remote_name, anames.remote_name, relation_name, anames.remote_field
                ),
            )
            .attach(
                QueryPart::Where,
                format!("`{}`.`{}` = ?", relation_name, anames.local_field),
            )
    }
    fn bind(&self, ctx: &mut StatementContext, index: &mut i32) {
        let rdata = self
            .get_data()
            .expect("binding query for relation with no data");

        ctx.bind(*index, rdata.local_id)
            .expect("couldn't bind relation id");
        *index += 1;
    }
}

impl<E: Entity, EPL: EntityPartList<Entity = E>> Index<E, EPL> {
    /// Perform a search through this index
    ///
    /// Note that this is simply sugar for `idmap_instance.indexed(self, values)`.
    pub fn search<'a>(
        &'a self,
        values: impl 'a + QueryEquivalentList<EPL::DatumList>,
    ) -> impl 'a + Queryable<EntityOutput = E, OutputContainer = Option<Stored<E>>> {
        self.indexed(self, values)
    }
}

impl<'a, E: Entity, EPL: EntityPartList<Entity = E>> Queryable for &'a Index<E, EPL> {
    type EntityOutput = E;
    type OutputContainer = Vec<Stored<E>>;
    type StaticVersion = &'static Index<E, EPL>;

    const IS_UNIQUE: bool = false;

    fn build(&self) -> Query {
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT")
            .attach(QueryPart::Columns, "*")
            .attach(QueryPart::From, format!("`{}`", E::entity_name()))
    }
    fn bind(&self, _stmt: &mut StatementContext, _index: &mut i32) {}
}
