//! Schema specification types.
//!
//! The following terminology used in some types in this module:
//! - *domain*: one side of a relation, the "pointed-from" side for one-sided relations
//! - *range*: one side of a relation, the "pointed-to" side for one-sided relations
//! - *local*: the current side of a relation
//! - *remote*: the opposite side of a relation

use crate::{
    db::{ConnectionLease, StatementContext, StatementRow},
    DBResult, Error,
};

use self::{
    datum::{ConcreteDatum, Datum, DatumDiscriminator, DatumDiscriminatorRef},
    entity::{Entity, EntityPartList},
};

/// Types related to datums, or struct fields.
pub mod datum;
/// Types related to entities, or structs that can be serialized into/deserialized from the
/// database.
pub mod entity;

/// Types related to inter-entity relationships.
pub mod relation;

/// Types related to indexes.
pub mod index;

mod build;
mod collect;
pub(crate) mod meta;

// ----------------------------------------------------------------------
// API types
// ----------------------------------------------------------------------

/// Wrapper struct that holds both an EntityID and an Entity itself.
pub struct Stored<T: Entity> {
    id: T::ID,
    wrap: T,
}

impl<T: Entity> Stored<T> {
    pub(crate) fn new(id: T::ID, value: T) -> Self {
        Self { id, wrap: value }
    }

    /// Retrieve the entity ID of the stored entity.
    pub fn id(&self) -> T::ID {
        self.id
    }

    /// Consume the `Stored<>` wrapper and return the wrapped data.
    pub fn wrapped(self) -> T {
        self.wrap
    }
}

impl<T: Entity + std::fmt::Debug> std::fmt::Debug for Stored<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "Stored {{ id: {:?}, value: {:?} }}",
            self.id, self.wrap
        ))
    }
}

impl<T: Entity> AsRef<T> for Stored<T> {
    fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> AsMut<T> for Stored<T> {
    fn as_mut(&mut self) -> &mut T {
        &mut self.wrap
    }
}

impl<T: Entity> std::ops::Deref for Stored<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrap
    }
}

impl<T: Entity> std::ops::DerefMut for Stored<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrap
    }
}

impl<T: Entity + Clone> Clone for Stored<T> {
    fn clone(&self) -> Self {
        Self {
            id: self.id,
            wrap: self.wrap.clone(),
        }
    }
}

impl<T: Entity> PartialEq for Stored<T> {
    fn eq(&self, other: &Self) -> bool {
        self.id == other.id
    }
}

// ----------------------------------------------------------------------
// Entity field types
// ----------------------------------------------------------------------

/// Stores an arbitrary Rust data type as serialized JSON in a string field.
#[derive(Clone)]
pub struct Serialized<T: serde::Serialize + serde::de::DeserializeOwned + Clone> {
    wrapped: T,
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> Serialized<T> {
    /// Returns the object stored inside this representational wrapper.
    pub fn wrapped(self) -> T {
        self.wrapped
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Default + Clone> Default
    for Serialized<T>
{
    fn default() -> Self {
        Self {
            wrapped: T::default(),
        }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone> std::fmt::Debug
    for Serialized<T>
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <T as std::fmt::Debug>::fmt(&self.wrapped, f)
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> From<T> for Serialized<T> {
    fn from(value: T) -> Self {
        Self { wrapped: value }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> AsRef<T> for Serialized<T> {
    fn as_ref(&self) -> &T {
        &self.wrapped
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> AsMut<T> for Serialized<T> {
    fn as_mut(&mut self) -> &mut T {
        &mut self.wrapped
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> std::ops::Deref for Serialized<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrapped
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Clone> std::ops::DerefMut
    for Serialized<T>
{
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrapped
    }
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone> Datum
    for Serialized<T>
{
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to(&self, stmt: &mut StatementContext, index: i32) {
        let json = std::pin::Pin::new(
            serde_json::to_string(&self.wrapped).expect("couldn't serialize object into JSON"),
        );

        <&str as Datum>::bind_to(&&*json.as_ref(), stmt, index);

        // transfer ownership so that the data is still around for later
        stmt.transfer(json);
    }

    fn build_from(
        rdata: relation::RelationData,
        stmt: &mut StatementRow,
        index: &mut i32,
    ) -> DBResult<Self>
    where
        Self: Sized,
    {
        let s = <String as Datum>::build_from(rdata, stmt, index)?;

        let d = serde_json::from_str::<T>(s.as_str()).map_err(Error::JSON)?;

        Ok(Self { wrapped: d })
    }

    fn accept_discriminator(d: &mut impl DatumDiscriminator)
    where
        Self: Sized,
    {
        d.visit_serialized::<T>();
    }

    fn accept_discriminator_ref(&self, d: &mut impl DatumDiscriminatorRef)
    where
        Self: Sized,
    {
        d.visit_serialized::<T>(&self.wrapped);
    }
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone>
    ConcreteDatum for Serialized<T>
{
}

/// Helper trait to make working with [`Serialized`] fields a little bit nicer.
pub trait Serializable:
    serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone
{
    /// Wrap an eligible object into a [`Serialized`] version of itself.
    fn into_serialized(self) -> Serialized<Self>
    where
        Self: Sized;
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug + Clone>
    Serializable for T
{
    fn into_serialized(self) -> Serialized<Self>
    where
        Self: Sized,
    {
        Serialized { wrapped: self }
    }
}

// ----------------------------------------------------------------------
// Database specification types
// ----------------------------------------------------------------------

/// Table with EntityID-based lookup.
pub struct IDMap<T: Entity> {
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> Default for IDMap<T> {
    fn default() -> Self {
        Self {
            _ghost: std::marker::PhantomData,
        }
    }
}

impl<T: Entity> Clone for IDMap<T> {
    fn clone(&self) -> Self {
        Self {
            _ghost: Default::default(),
        }
    }
}

impl<T: Entity> IDMap<T> {
    /// Construct a non-empty instance of an `IDMap`.
    pub fn build() -> Self {
        Self {
            _ghost: std::marker::PhantomData,
        }
    }
}

impl<E: Entity> DatabaseItem for IDMap<E> {
    fn accept_item_visitor(&self, visitor: &mut impl DatabaseItemVisitor) {
        visitor.visit_idmap::<E>();
    }
}

/// Represents a logical item in a database schema, be it an index, table, or logical grouping.
pub trait DatabaseItem {
    /// Accept an entity visitor for entity discovery.
    fn accept_item_visitor(&self, visitor: &mut impl DatabaseItemVisitor);
}

/// Visitor trait for iterating across the types in a [`Schema`] tree.
pub trait DatabaseItemVisitor {
    /// Visit an `IDMap<T>` type.
    fn visit_idmap<T: Entity>(&mut self)
    where
        Self: Sized;
    /// Visit an `Index<T, PL>` type.
    fn visit_index<T: Entity, PL: EntityPartList<Entity = T>>(&mut self)
    where
        Self: Sized;
}

/// A root structure for the database specification graph.
pub trait Schema {
    /// Install this schema into a database
    fn install(&self, lease: &mut ConnectionLease) -> DBResult<()>
    where
        Self: Sized,
    {
        let schema = build::collect_from_database(self);
        match schema.check(lease) {
            // schema checks out
            Some(true) => {},
            // schema doesn't match
            Some(false) => Err(Error::IncompatibleSchema)?,
            // no schema found
            None => {
                schema.create(lease)?;
            },
        }
        Ok(())
    }

    /// Accept a visitor for iteration.
    fn accept_item_visitor(&self, visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized;
}

impl<S: Schema> DatabaseItem for S {
    fn accept_item_visitor(&self, visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized,
    {
        <Self as Schema>::accept_item_visitor(self, visitor);
    }
}
