use std::collections::HashMap;

use crate::{
    db::{ConnectionLease, Transaction},
    query::{Insertable, Queryable},
    schema::{
        collect::{EntityStateContainer, PartType},
        entity::{Entity, EntityPart, EntityPartList, EntityPartVisitor},
        meta, DatabaseItemVisitor,
    },
    DBResult,
};

use sha2::Digest;

use super::Schema;

#[derive(Debug)]
struct ColumnInfo {
    name: &'static str,
    ty: String,
    fkey: Option<String>,
    unique: bool,
}

#[derive(Debug)]
struct TableInfo {
    table_name: String,
    columns: Vec<ColumnInfo>,
    constraints: Vec<String>,
}

impl TableInfo {
    fn new(name: String) -> Self {
        TableInfo {
            table_name: name,
            columns: vec![],
            constraints: vec![],
        }
    }

    fn build_creation_query(&self) -> String {
        let columns = self.columns.iter().map(|col| {
            format!(
                ", `{}` {}{}",
                col.name,
                col.ty,
                if col.unique { " unique" } else { "" }
            )
        });
        let fkeys = self.columns.iter().filter_map(|col| {
            Some(format!(
                ", foreign key(`{}`) references {} on delete cascade",
                col.name,
                col.fkey.as_ref()?
            ))
        });

        format!(
            "create table `{}` (`id` integer primary key{}{}{});",
            self.table_name,
            columns.collect::<String>(),
            fkeys.collect::<String>(),
            self.constraints
                .iter()
                .fold(String::new(), |a, b| format!("{}, {}", a, b))
        )
    }
}

#[derive(Debug, Hash)]
struct IndexInfo {
    table_name: String,
    columns: Vec<String>,
    unique: bool,
}

impl IndexInfo {
    fn build_creation_query(&self) -> String {
        let index_name = format!(
            "index_{}_{}",
            self.table_name,
            self.columns
                .iter()
                .cloned()
                .reduce(|a, b| format!("{a}_{b}"))
                .unwrap()
        );

        format!(
            "create {unique} index `{index_name}` on `{table_name}`({cols})",
            table_name = self.table_name,
            unique = if self.unique { "unique" } else { "" },
            cols = self.columns.join(",")
        )
    }
}

#[derive(Debug)]
pub(crate) struct DatabaseSchema {
    signature: String,
    table_queries: HashMap<String, String>,
    index_queries: HashMap<String, String>,
}

impl DatabaseSchema {
    pub(crate) const SCHEMA_SIGNATURE_KEY: &'static str = "schema_signature";

    // the following functions are for use in migrations
    #[allow(unused)]
    pub(crate) fn signature(&self) -> &str {
        self.signature.as_str()
    }

    #[allow(unused)]
    pub(crate) fn table_queries(&self) -> &HashMap<String, String> {
        &self.table_queries
    }

    #[allow(unused)]
    pub(crate) fn index_queries(&self) -> &HashMap<String, String> {
        &self.index_queries
    }

    /// Three possible results:
    /// - yes, this is a schema match (true)
    /// - no, this is not a schema match (false)
    /// - there is no schema that we know of (None)
    pub fn check(&self, lease: &mut ConnectionLease) -> Option<bool> {
        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::default();

        // check to see if the signature exists and matches
        metadb
            .metastore
            .keyed(&Self::SCHEMA_SIGNATURE_KEY.to_string())
            .get(lease)
            .ok()
            .flatten()
            .map(|kv| kv.value == self.signature)
        // .map(|kv| kv.value.parse::<u64>().unwrap_or(0) == self.signature)
    }

    pub fn create(&self, lease: &mut ConnectionLease) -> DBResult<()> {
        let mut txn = Transaction::new(lease, "_microrm_create")?;
        for query in self.table_queries.iter() {
            log::trace!("Running creation query {}", query.1);
            txn.lease().execute_raw_sql(query.1)?;
        }
        for query in self.index_queries.iter() {
            log::trace!("Running creation query {}", query.1);
            txn.lease().execute_raw_sql(query.1)?;
        }

        // attempt to use connection as a MetadataDB database
        let metadb = meta::MetadataDB::default();

        for query in collect_from_database(&meta::MetadataDB::default())
            .table_queries
            .iter()
        {
            txn.lease().execute_raw_sql(query.1)?;
        }

        // store signature
        metadb.metastore.insert(
            txn.lease(),
            meta::MicrormMeta {
                key: Self::SCHEMA_SIGNATURE_KEY.into(),
                value: self.signature.clone(),
            },
        )?;

        txn.commit()?;

        Ok(())
    }
}

pub(crate) fn collect_from_database<DB: Schema>(schema: &DB) -> DatabaseSchema {
    struct IV(EntityStateContainer, Vec<IndexInfo>);

    impl DatabaseItemVisitor for IV {
        fn visit_idmap<E: Entity>(&mut self)
        where
            Self: Sized,
        {
            self.0.visit_idmap::<E>();
        }

        fn visit_index<T: Entity, PL: EntityPartList<Entity = T>>(&mut self)
        where
            Self: Sized,
        {
            struct PV<E: Entity>(Vec<String>, std::marker::PhantomData<E>);
            impl<E: Entity> EntityPartVisitor for PV<E> {
                type Entity = E;

                fn visit<EP: EntityPart<Entity = Self::Entity>>(&mut self) {
                    self.0.push(EP::part_name().to_string());
                }
            }
            let mut pv = PV::<T>(Default::default(), Default::default());
            PL::accept_part_visitor(&mut pv);

            self.1.push(IndexInfo {
                table_name: T::entity_name().to_string(),
                columns: pv.0,
                unique: true,
            });
        }
    }

    let mut iv = IV(EntityStateContainer::default(), Default::default());

    schema.accept_item_visitor(&mut iv);

    // now to turn all that into a set of tables
    let mut tables = std::collections::HashMap::new();

    for state in iv.0.iter_states() {
        let table_name = state.name.to_string();
        // we may end up visiting duplicate entities; skip them if so
        if tables.contains_key(&table_name) {
            continue;
        }

        let mut table = TableInfo::new(table_name.clone());
        for part in state.parts.iter() {
            match &part.ty {
                PartType::Datum(dtype) => table.columns.push(ColumnInfo {
                    name: part.name,
                    ty: dtype.to_string(),
                    fkey: None,
                    unique: part.unique,
                }),
                PartType::IDReference(entity_name) => table.columns.push(ColumnInfo {
                    name: part.name,
                    ty: "int".into(),
                    fkey: Some(format!("`{}`(`id`)", entity_name)),
                    unique: part.unique,
                }),
                PartType::RelationDomain {
                    table_name: relation_table_name,
                    range_name,
                    injective,
                } => {
                    let mut relation_table = TableInfo::new(relation_table_name.clone());

                    relation_table.columns.push(ColumnInfo {
                        name: "domain",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", table_name)),
                        unique: false,
                    });

                    relation_table.columns.push(ColumnInfo {
                        name: "range",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", range_name)),
                        unique: *injective,
                    });

                    relation_table
                        .constraints
                        .push("unique(`range`, `domain`)".to_string());

                    tables.insert(relation_table_name.clone(), relation_table);
                },
                PartType::RelationRange {
                    table_name: relation_table_name,
                    domain_name,
                    injective,
                } => {
                    let mut relation_table = TableInfo::new(relation_table_name.clone());

                    relation_table.columns.push(ColumnInfo {
                        name: "domain",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", domain_name)),
                        unique: false,
                    });

                    relation_table.columns.push(ColumnInfo {
                        name: "range",
                        ty: "int".into(),
                        fkey: Some(format!("`{}`(`id`)", table_name)),
                        unique: *injective,
                    });

                    relation_table
                        .constraints
                        .push("unique(`range`, `domain`)".to_string());
                    tables.insert(relation_table_name.clone(), relation_table);
                },
            }
        }

        let key = state.parts.iter().filter(|p| p.key).collect::<Vec<_>>();
        if !key.is_empty() {
            table.constraints.push(format!(
                "/* keying index */ unique({})",
                key.into_iter()
                    .map(|s| format!("`{}`", s.name))
                    .reduce(|a, b| format!("{},{}", a, b))
                    .unwrap()
            ));
        }

        tables.insert(table_name, table);
    }

    // this must be a stable hash function, so we use sha2
    let mut hasher = sha2::Sha256::new();

    let mut table_queries = vec![];
    let mut index_queries = vec![];

    // create sorted table list
    let mut sorted_tables: Vec<_> = tables.into_iter().collect();
    sorted_tables.sort_by(|a, b| a.0.cmp(&b.0));

    for (table_name, table) in sorted_tables {
        let create_sql = table.build_creation_query();

        hasher.update(&table_name);
        hasher.update(&create_sql);
        table_queries.push((table_name, create_sql));
    }

    for iinfo in iv.1.into_iter() {
        hasher.update(&iinfo.table_name);
        let query = iinfo.build_creation_query();
        hasher.update(&query);
        index_queries.push((iinfo.table_name, query));
    }

    let digest = hasher.finalize();

    DatabaseSchema {
        signature: digest.into_iter().fold(String::new(), |mut a, v| {
            a += &format!("{:02x}", v);
            a
        }),
        table_queries: HashMap::from_iter(table_queries),
        index_queries: HashMap::from_iter(index_queries),
    }
}
